;; $Id: ch6-base.scm,v 1.35 1999/04/29 17:25:45 leavens Exp $

;; Code from the EOPL chapter 6 revision, Figure 6.1.2
;; augmented with the local feature of Figures 6.4.1 and 6.5.1
;; and with the letproc feature of Figure 6.4.1.

;; To this must be added code that determines how parameters are handled
;; (such as figure 6.1.3 or a variation).

;;; Grammar:
;;         <form> ::= <exp> | define <var> = exp | definearray <var> <exp>
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;                  | if <exp> then <exp> else <exp>
;;	            | proc <varlist> <exp>
;;	            | <var> := <exp>
;;                  | begin <exp> {; <exp>}* end
;;                  | letarray <arraydecls> in <exp>
;;		    | <array-exp> [<exp>]
;;	            | <array-exp> [<exp>] := <exp>
;;                  | letproc <procdecls> in <exp>
;;                  | local <decls> in <exp>
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( ) | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*
;;	   <array-exp> ::= <varref> | (<exp>)
;;	  <arraydecls> ::= <arraydecl> {; <arraydecl>}*
;;	   <arraydecl> ::= <var> [<exp>]
;;         <procdecls> ::= <procdecl> {; <procdecl>}*
;;	 <procdecl> ::= <var> <varlist> = <exp>

;;; Parsing

;; To use "run", you also need to do
;;        (load "run.scm")
;; To use "read-eval-print", you also need to do
;;        (load "rep.scm")
;; Note that you can't use both run and read-eval-print at the same time.

(load-quietly-from-lib "undefine-read-eval-print.scm")
(load-quietly-from-lib "undefine-run.scm")
(load-quietly-from-lib "ch6-char-parser.scm")

;;; Syntactic sugars
;;       see the file below

(load-quietly-from-lib "syntax-expand-letrec.scm")

;;; Top-level Forms
;;;      see the file below

(load-quietly-from-lib "eval-form-with-define.scm")

;;; ADTs used

(load-quietly-from-lib "cell.scm")
(load-quietly-from-lib "updateable-environments.scm")
(load-quietly-from-lib "procedure.scm")
(load-quietly-from-lib "array.scm")

;;; Helpful stuff

(load-quietly-from-lib "compose.scm")
(load-quietly-from-lib "ignore.scm")

;;; Truth values

;; Anonymous figure : page 146

(define true-value?
  ; TYPE: (-> (integer) boolean)
  (lambda (x)
    (not (zero? x))))

(define boolean->int
  ;; TYPE: (-> (boolean) integer)
  (lambda (x)
    (if  x  1  0)))

;;; To help the type checker, we load some set of the domains here.
;;; These will be overridden ...
(load-quietly-from-lib "ch6-1-3-denoted-value.scm")
(load-quietly-from-lib "ch6-1-3-expressed-value.scm")

;;; The following procedures are defined here again, just to help the type checker
;;; They will also be overridden ...
;;; Figure 6.1.3 : page 183

(define denoted->expressed
  ;; TYPE: (-> (Denoted-Value) Expressed-Value)
  cell-ref)

(define expressed->denoted
  ;; TYPE: (-> (Expressed-Value) Denoted-Value)
  make-cell)

(define denoted-value-assign!
  ;; TYPE: (-> (Denoted-Value Expressed-Value) void)
  cell-set!)

(define do-letarray
  ;; TYPE: (-> (number) Denoted-Value)
  (lambda (n)
    (if (not (<= 0 n))  ; added error check vs. text
	(error "do-letarray: bad index")
	(make-cell (array->expressed (make-array n))))))

(define eval-array-exp
  ;; TYPE: (-> (parsed-exp Environment) Expressed-Value)
  (lambda (exp env)
    (eval-exp exp env)))

;; addition to figure 6.1.3

(define array-element-assign!
  ;; TYPE: (-> ((Array Expressed-Value) number Expressed-Value) void)
  array-set!)


;;; Evaluation of expressions, common to all the interpreters in this chapter

;;; Figure 6.1.2 : page 182

(define eval-exp
  ; TYPE: (-> (parsed-exp Environment) Expressed-Value)
  (lambda (exp env)
    (variant-case exp
      (lit (datum) (number->expressed datum))
      (varref (var) (denoted->expressed (apply-env env var)))
      (app (rator rands) 
        (apply-proc (expressed->procedure (eval-rator rator env))
		    (eval-rands rands env)))
      (if (test-exp then-exp else-exp)
	  (if (true-value? (expressed->number (eval-exp test-exp env)))
	      (eval-exp then-exp env)
	      (eval-exp else-exp env)))
      (proc (formals body) 
	    (procedure->expressed (make-closure formals body env)))
      (varassign (var exp)
        (void->expressed (denoted-value-assign! (apply-env env var)
						(eval-exp exp env))))
      (letarray (arraydecls body)
        (eval-exp body 
          (extend-env
	    (map decl->var arraydecls)
            (map (lambda (decl)
                   (do-letarray
		    (expressed->number (eval-exp (decl->exp decl) env))))
                 arraydecls)
            env)))
      (arrayref (array index) 
        (array-ref (expressed->array (eval-array-exp array env))
		   (expressed->number (eval-exp index env))))
      (arrayassign (array index exp)
	(void->expressed
	 (array-element-assign!        ;; changed from the revised chapter 6
	  (expressed->array (eval-array-exp array env))
          (expressed->number (eval-exp index env))
          (eval-exp exp env))))
      (begin (exp1 exp2)
	(begin
	  (ignore (eval-exp exp1 env))
	  (eval-exp exp2 env)))
      ;; letproc and local are for sections 6.4 and 6.5 (see Figure 6.4.1)
      (letproc (procdecls body)
        (let ((vars (map procdecl->var procdecls))
              (closures (map (lambda (decl)
			       (expressed->denoted
				(procedure->expressed
				 (make-closure
				  (procdecl->formals decl)
				  (procdecl->body decl)
				  env))))
			     procdecls)))
          (let ((new-env (extend-env vars closures env)))
            (eval-exp body new-env))))
      (local (decls body)
        (let ((vars (map decl->var decls))
              (exps (map decl->exp decls)))
          (let ((new-env (extend-env
			   vars
                           (map (lambda (exp) 
                                  (expressed->denoted   ;; changed!
				    (eval-exp exp env)))
                                exps)
                           env)))
            (eval-exp body new-env))))
      (else (error "Invalid abstract syntax: " exp)))))

(define eval-rator
  ; TYPE: (-> (parsed-exp Environment) Expressed-Value)
  (lambda (rator env)
    (eval-exp rator env)))

(define eval-rands
  ; TYPE: (-> ((list parsed-exp) Environment) (list Denoted-Value))
  (lambda (rands env)
    (map (lambda (rand) (eval-rand rand env)) rands)))

(define eval-rand
  ; TYPE: (-> (parsed-exp Environment) Denoted-Value)
  (lambda (exp env)
    (expressed->denoted (eval-exp exp env))))

(define apply-proc
  ; TYPE: (-> (Procedure (list Denoted-Value)) Expressed-Value)
  (lambda (proc args)
    (variant-case proc
      (prim-proc (prim-op)
		 (apply-prim-op prim-op (map denoted->expressed args)))
      (closure (formals body env)
	       (if (= (length formals) (length args))
		   (eval-exp body (extend-env formals args env))
		   (error "Wrong number of arguments passed to closure"
			  "\n      whose body is " body
			  "\n      expecting " (length formals) " arguments"
			           " but got " (length args))))
      (else (error "Invalid procedure:" proc)))))


(define apply-prim-op
  ;; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (number->expressed (+ (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((*) (number->expressed (* (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((-) (number->expressed (- (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((add1) (number->expressed (+ (expressed->number (car args)) 1)) )
      ((sub1) (number->expressed (- (expressed->number (car args)) 1)) )
      ;; primitive operators added to interpreter follow
      ((minus) (number->expressed (- (expressed->number (car args)))))
      ((cons) (list->expressed (cons (car args)
				     (expressed->list (cadr args)))))
      ((car) (car (expressed->list (car args))))
      ((cdr) (list->expressed (cdr (expressed->list (car args)))))
      ((cadr) (cadr (expressed->list (car args))))
      ((caddr) (caddr (expressed->list (car args))))
      ((cadddr) (cadddr (expressed->list (car args))))
      ((list) (list->expressed args))
      ((null) (number->expressed (boolean->int
				  (null? (expressed->list (car args))))) )
      ((eq) (number->expressed (boolean->int (eq? (car args) (cadr args)))) )
      ((equal) (number->expressed (boolean->int
				   (= (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((less) (number->expressed (boolean->int
				   (< (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((greater) (number->expressed (boolean->int
				   (> (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((zero) (number->expressed (boolean->int
				  (zero? (expressed->number (car args))))) )
      ((write) (void->expressed (write (car args))))
      ((print) (void->expressed (apply displayln args)))
      ((writeln) (void->expressed (begin (for-each (lambda (arg)
						     (write arg)
						     (display " "))
						   args)
					 (newline))))
      ((space) (void->expressed (display " ")))
      ((newline) (void->expressed (newline)))
      ((printchar) (void->expressed
		    (display (integer->char
			      (expressed->number (car args))))))
      ((arraylength) (number->expressed (array-length (expressed->array (car args)))))
      (else (error "Invalid prim-op name:" prim-op)))))

(define prim-op-names  ; TYPE: (list symbol)
  ;; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr write print writeln space newline
      printchar arraylength))

(define init-env       ; TYPE: Environment
  (extend-env
    prim-op-names
    (map (compose expressed->denoted
		  (compose procedure->expressed make-prim-proc))
	 prim-op-names)
    ;; modified to add the empty list
    (extend-env
      '(emptylist)
      (list (expressed->denoted (list->expressed '())))
      the-empty-env)))
