;;; $Id: combinator-tools.scm,v 1.3 1999/02/11 06:25:19 leavens Exp $

;;; The following algorithm is from Hindley, Lercher and Seldin's book
;;;     "Introduction to Combinatory Logic", Cambridge, 1972)

;;; For the input grammar, <c-lambda-1-exp>, see the following file:
(load-quietly "combinator-lambda-1.scm")

;;; For the output grammar, <combinator-term>, see the following file:
(load-quietly "combinator-term.scm")

;;; Examples:
;;; (unparse-combinator-term
;;;   (toCombinators (parse-c-lambda-1-exp '(lambda (x) (lambda (y) x)))))
;;;  ==> k
;;; (unparse-combinator-term
;;;   (toCombinators (parse-c-lambda-1-exp '(lambda (x) (lambda (y) y)))))
;;;  ==> (k i)

(define toCombinators
  ;; TYPE: (-> (c-lambda-1-exp) combinator-term)
  (lambda (trm)
    (variant-case trm
      (varref (var) (make-var var))
      (app (rator rand) (make-capp (toCombinators rator) (toCombinators rand)))
      (lambda (formal body)
	(if (not (occurs-free-in? formal body))
	    (make-capp (make-K) (toCombinators body))
	    (let ((outer-formal formal))
	      (variant-case body
		(varref (var) (make-I))   ; because formal == var here
		(lambda ()
		  (toCombinators
		   (make-lambda outer-formal
				(toLambda (toCombinators body)))))
		(app (rator rand)
		     (if (and (equal? rand (make-varref outer-formal))
			      (not (occurs-free-in? outer-formal rator)))
			 (toCombinators rator)  ; eta rule
			 (make-capp
			  (make-capp (make-S)
				     (toCombinators
				      (make-lambda outer-formal rator)))
			  (toCombinators (make-lambda outer-formal rand)))))
		(else (toCombinators body))))))
      (LI () (make-I))
      (LS () (make-S))
      (LK () (make-K))
      (else (error "bad c-lambda-1-exp term: " trm)))))

(define interpret
  ;; TYPE: (-> (combinator-term) combinator-term)
  (lambda (trm)
    ;; ENSURES LIBERALLY: result is the normal form of trm
    (displayln (unparse-combinator-term trm) " ==> ")
    (cond
     ((capp-of-I? trm) (interpret (capp->rand trm)))
     ((capp-of-K? trm) (interpret (capp->rand (capp->rator trm))))
     ((capp-of-S? trm)
      (interpret (make-capp
		  (make-capp (capp->rand (capp->rator (capp->rator trm)))
			     (capp->rand trm))
		  (make-capp (capp->rand (capp->rator trm))
			     (capp->rand trm)))))
     (else trm))))

(define capp-of-I?
  ;; TYPE: (-> (combinator-term) boolean)
  (lambda (trm)
    (and (capp? trm)
	 (I? (capp->rator trm)))))

(define capp-of-K?
  ;; TYPE: (-> (combinator-term) boolean)
  (lambda (trm)
    (and (capp? trm)
	 (capp? (capp->rator trm))
	 (K? (capp->rator (capp->rator trm))))))

(define capp-of-S?
  ;; TYPE: (-> (combinator-term) boolean)
  (lambda (trm)
    (and (capp? trm)
	 (capp? (capp->rator trm))
	 (capp? (capp->rator (capp->rator trm)))
	 (S? (capp->rator (capp->rator (capp->rator trm)))))))


(define lambda-calculus-rep
  ;; TYPE: (-> () poof)
  (lambda ()
    ;; EFFECT: read, interpret, and print lambda calculus expressions.
    ;;         You can enter combinator terms directly by using LS, LK, and LI.
    ;;         You can quit by typing in something that doesn't parse, like
    ;;            (quit)
    (display "lambda-calculus> ")
    (force-output)   ; SCM specific
    (pretty-print
     (unparse-combinator-term
      (interpret
       (toCombinators
	(parse-c-lambda-1-exp
	 (read))))))
    (newline)
    (lambda-calculus-rep)))

;;; The toLambda function is used internally above;
;;; it does not attempt to recover the initial forma of a lambda term,
;;; which would (in general) be impossible anyway.

(define toLambda
  ;; TYPE: (-> (combinator-term) c-lambda-1-exp)
  (lambda (ctrm)
    (variant-case ctrm
      (I () (make-LI))
      (K () (make-LK))
      (S () (make-LS))
      (var (symbol) (make-varref symbol))
      (capp (rator rand) (make-app (toLambda rator) (toLambda rand)))
      (else (error "Bad term given to toLambda: " ctrm)))))

(define occurs-free-in?
  ;; TYPE: (-> (symbol c-lambda-1-exp) boolean)
  (lambda (sym trm)
    ;; ENSURES: result is true just when sym occurs free in trm
    (variant-case trm
     (varref (var) (eq? sym var))
     (lambda (formal body)
       (and (not (eq? sym formal)) (occurs-free-in? sym body)))
     (app (rator rand)
	  (or (occurs-free-in? sym rator)
	      (occurs-free-in? sym rand)))
     (else #f))))
