;;; $Id: lexical-addr-exp.scm,v 1.5 1999/02/08 21:30:11 leavens Exp $
;;; An ADT for lexical address expressions; this is essentially
;;; the untyped lambda multiple-argument calculus
;;; with quoted symbols and if expressions, but with lexical addresses
;;; replacing varrefs.

;;;	<lexical-addr-exp> ::= <lexical-addr>
;;;               | ( quote <symbol> )
;;;               | (lambda ({<var>}*) <lexical-addr-exp>)
;;;   	          | (if <lexical-addr-exp> <lexical-addr-exp>
;;;                                        <lexical-addr-exp>)
;;;               | (<lexical-addr-exp> {<lexical-addr-exp>}*)
;;;		  | ( lambda ( {<var>}* ) <lcm+if-exp> )
;;;     <lexical-addr> ::= ( <symbol> : <number> <number> )
;;;     <var> ::= <symbol>
;;;     <varref> ::= <symbol>

(load-quietly-from-lib "all.scm")
(load-from-lib "lexical-addr.scm")

(define la:lexical-addr-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (exp)
    (or (la:varref? exp)
	(and (la:quote? exp) 
	     (symbol? (la:quote->symbol exp)))
	(and (la:lambda? exp) 
	     (la:lexical-addr-exp? (la:lambda->body exp)))
	(and (la:if? exp)
	     (la:lexical-addr-exp? (la:if->test-exp exp))
	     (la:lexical-addr-exp? (la:if->then-exp exp))
	     (la:lexical-addr-exp? (la:if->else-exp exp)))
	(and (la:app? exp)
	     (la:lexical-addr-exp? (la:app->rator exp))
	     (all la:lexical-addr-exp? (la:app->rands exp))))))

(define la:varref? lexical-addr?)  ; TYPE: (-> (lexical-addr-exp) boolean)

(define la:quote?   ; TYPE: (-> (lexical-addr-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 2)
	 (eq? (car exp) 'quote)
	 (symbol? (cadr exp)))))

(define la:lambda?  ; TYPE: (-> (lexical-addr-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'lambda)
	 (let ((formals (cadr exp)))
	   (and (list? formals)
		(all symbol? formals))))))

(define la:if?  ; TYPE: (-> (lexical-addr-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 4)
	 (eq? (car exp) 'if))))

(define la:app?  ; TYPE: (-> (lexical-addr-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (>= (length exp) 1)
	 (let ((operator (car exp)))
	   (and (not (eq? operator 'quote))
		(not (eq? operator 'lambda))
		(not (eq? operator 'if)))))))

(define la:varref->lexical-addr  ; TYPE: (-> (lexical-addr-exp) lexical-addr)
  ;; REQUIRES: the argument is a lexical-addr
  (lambda (vr) vr))

(define la:quote->symbol   ; TYPE: (-> (lexical-addr-exp) symbol)
  ;; REQUIRES: the argument is a quoted expression
  cadr)

(define la:lambda->formals   ; TYPE: (-> (lexical-addr-exp) (list symbol))
  ;; REQUIRES: the argument is a lambda
  (lambda (exp)
    (cadr exp)))

(define la:lambda->body   ; TYPE: (-> (lexical-addr-exp) lexical-addr-exp)
  ;; REQUIRES: the argument is a lambda
  (lambda (exp)
    (caddr exp)))

(define la:app->rator   ; TYPE: (-> (lexical-addr-exp) lexical-addr-exp)
  ;; REQUIRES: the argument is an application
  (lambda (exp)
    (car exp)))

(define la:app->rands    ; TYPE: (-> (lexical-addr-exp) (list lexical-addr-exp))
  ;; REQUIRES: the argument is an application expression
  (lambda (exp)
    (cdr exp)))

(define la:if->test-exp
  ; TYPE: (-> (lexical-addr-exp) lexical-addr-exp)
  cadr)

(define la:if->then-exp
  ; TYPE: (-> (lexical-addr-exp) lexical-addr-exp)
  caddr)

(define la:if->else-exp
  ; TYPE: (-> (lexical-addr-exp) lexical-addr-exp)
  cadddr)

(define la:make-varref
  ;; TYPE: (-> (lexical-addr) lexical-addr-exp)
  (lambda (la)
    la))

(define la:make-quote
  ;; TYPE: (-> (symbol) lexical-addr-exp)
  (lambda (sym)
    (list 'quote sym)))

(define la:make-lambda
  ;; TYPE: (-> ((list symbol) lexical-addr-exp) lexical-addr-exp)
  (lambda (formals body)
    (list 'lambda formals body)))

(define la:make-if
  ;; TYPE: (-> (lexical-addr-exp lexical-addr-exp lexical-addr-exp) lexical-addr-exp)
  (lambda (test consequent alternative)
    (list 'if test consequent alternative)))

(define la:make-app
  ;; TYPE: (-> (lexical-addr-exp (list lexical-addr-exp)) lexical-addr-exp)
  cons)
  
(define la:parse-lexical-addr-exp
  ;; TYPE: (-> (datum) lexical-addr-exp)
  (lambda (exp)
    (cond
     ((lexical-addr? exp) (la:make-varref (parse-lexical-addr exp)))
     ((la:quote? exp) (la:make-quote (cadr exp)))
     ((la:lambda? exp)
      (la:make-lambda (cadr exp) (la:parse-lexical-addr-exp (caddr exp))))
     ((la:if? exp) (la:make-if (la:parse-lexical-addr-exp (cadr exp))
			    (la:parse-lexical-addr-exp (caddr exp))
			    (la:parse-lexical-addr-exp (cadddr exp))))
     ((la:app? exp) (la:make-app (la:parse-lexical-addr-exp (car exp))
			      (map la:parse-lexical-addr-exp (cdr exp))))
     (else (error "la:parse-lexical-addr-exp: bad syntax in expression: "
		  exp)))))
