;;; $Id: type-check-env.scm,v 1.8 1999/06/23 16:57:22 leavens Exp $
;;; Environments for the Scheme tc:type checker.
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.

;;;----------------------------------------------------------------------------
;;;
;;; An environment is an object that associates variables to values.
;;; Environments are represented by association lists (i.e., lists of pairs)
;;; (see tc:type-check-env.def).

(define tc:env?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (and (list? d)
	 (tc:all (lambda (b) (pair? b)) d))))

(define tc:env-empty 
  ;; TYPE: (-> () (env V T))
  (lambda ()
    ;; ENSURES: result is an empty environment
    '()))

(define tc:env-empty?
  ;; TYPE: (-> ((env V T)) boolean)
  null?)

(define tc:env-update 
  ;; TYPE: (-> (V T (env V T)) (env V T))
  (lambda (var val env)
    ;; ENSURES: result is a copy of env but with var associated to val
    (cons (cons var val) env)))

(define tc:env-update*
  ;; TYPE: (-> ((list V) (list T) (env V T)) (env V T))
  (letrec ((env-update*
	    (lambda (lvar lval)
	      ;; TYPE (-> ((list V) (list T)) (env V T))
	      (if (null? lvar)
		  (tc:env-empty)
		  (tc:env-update (car lvar)
				 (car lval)
				 (env-update* (cdr lvar) (cdr lval)))))))
    (lambda (lvar lval env)
      ;; EFFECT: result is a new environment binding each element of lvar to
      ;;         the corresponding element of lval and containing the bindings
      ;;         in env.  Note that the new bindings have precedence over 
      ;;         those in env and the new bindings are in the list order. Note
      ;;         that this procedure would be more efficiently implemented
      ;;         using tail-recursion but that this would reverse the order of
      ;;         the new bindings.
      (tc:env-join (env-update* lvar lval) env))))

(define tc:env-join  
  ;; TYPE: (-> ((env V T) (env V T)) (env V T))
  (lambda (env1 env2)
    ;; ENSURES: result is a new environment containing the bindings
    ;; in env1 and env2 (env1 has precedence)
    (append env1 env2)))

(define tc:env-bound? 
  ;; TYPE: (-> (V (env V T)) boolean)
  (lambda (var env)
    ;; ENSURES: result is true iff var is bound in env
    (let ((res (assq var env)))
      (if (and res (not (pair? res)))
	  (error "funny result from tc:env-bound?: " res "for " env)
	  res))))

(define tc:env-value  
  ;; TYPE: (-> (V (env V T)) T)
  (lambda (var env)
    ;; REQUIRES: var has an association in env
    ;; ENSURES: result is the value associated to var in env
    (cdr (assq var env))))

(define tc:env-map
  ;; TYPE: (-> ((-> (pair V T) (pair U S)) (env V T)) (env U S))
  (lambda (f env)
    ;; EFFECT: returns an environment that is like env but with
    ;;         every binding in env replaced with the result of
    ;;         f on the binding.
    (map f env)))

(define tc:env-first
  ;; TYPE: (-> ((env V T)) (pair V T))
  car)

(define tc:env-rest
  ;; TYPE: (-> ((env V T)) (env V T))
  cdr)

;;; Gary: I added this for assistance in converting the global var types
;;;       to environments

(define tc:assoc-list->env
  ;; TYPE: (-> ((list (pair V T))) (env V T))
  (lambda (x) x))


