;;; $Id: type-check-unify.scm,v 1.15 1999/06/23 16:57:22 leavens Exp $
;;; SYNOPSIS: unification for type expressions
;;; Code from "Paradigms of Artificial Intelligence Programming"
;;; by Peter Norvig, adapted to Scheme by Gary T. Leavens
;;; with some modifications by Steven L. Jenkins and Curtis Clifton
;;; Rewritten to be typeable and to take into account and-types
;;; by Gary T. Leavens.

(load-quietly-from-lib "maybe.scm")
(load-quietly-from-lib "type-check-env.scm")
(load-quietly-from-lib "type-check-type-expr.scm")

;-----------------------------------------------------------------------------

(define *tc:unification-failure* ; TYPE: (maybe (env uniq-sym type-expr))
  (make-nothing))

(define tc:unification-error?
  ;; TYPE: (-> (datum) boolean)
  (lambda (res)
    (maybe-nothing? res)))

;-----------------------------------------------------------------------------
(define tc:unify 
  ;; TYPE: (-> (type-expr
  ;;            type-expr
  ;;            (env uniq-sym type-expr)
  ;;            (list variable-type-expr))
  ;;           (maybe (env uniq-sym type-expr)))
  (lambda (x y bindings var-list)
    ;; ENSURES: result is either *tc:unification-failure*
    ;;          or the env that unifies x and y
    ;; Note: the ordering is important below.  In particular variables have to
    ;;       be handled before poof and datum, so that they will get bound
    ;;       to poof or datum as needed, and and-types have to be handled
    ;;       before ... types so that the common case for overload functions
    ;;       with ... in argument types works.
    (if (>= *tc:tracing-level* 3)
	(begin
	  (displayln "tc:unify")
	  (displayln "   x = " x)
	  (displayln "   y = " y)
	  (displayln "   bindings = " bindings)
	  (displayln "   var-list = " var-list)))
    ;; (if (not (tc:type-expr? x))
    ;;     (error "tc:unify called with non-type argument in x " x))
    ;; (if (not (tc:type-expr? y))
    ;;	   (error "tc:unify called with non-type argument in y " y))
    ;; (if (not (tc:env? bindings))
    ;;     (error "tc:unify called with non-env argument in bindings "
    ;;            bindings))
    (cond
     ((equal? x y) (make-something bindings))
     ((tc:variable-type-expr? x)
      (tc:unify-variable (tc:variable-type-expr->uniq-sym x)
			 y
			 bindings
			 var-list))
     ((tc:variable-type-expr? y)
      (tc:unify-variable (tc:variable-type-expr->uniq-sym y)
			 x
			 bindings
			 var-list))
     ((and (not (tc:error-type? x))
	   (not (tc:error-type? y))
	   (or (tc:poof-type-expr? x) (tc:poof-type-expr? y)))
      (make-something bindings))
     ((and (not (tc:error-type? x))
	   (not (tc:error-type? y))
	   (or (tc:datum-type-expr? x) (tc:datum-type-expr? y)))
      (make-something bindings))
     ((and (tc:basic-type-expr? x) (tc:basic-type-expr? y))
      (if (tc:basic-type-expr-eq? x y)
	  (make-something bindings)
	  *tc:unification-failure*))
     ((and (tc:applied-type-expr? x) (tc:applied-type-expr? y))
      (if (not
	   (tc:basic-type-expr-eq? (tc:applied-type-expr->operator-type x)
				   (tc:applied-type-expr->operator-type y)))
	  *tc:unification-failure*
	  (tc:list-unify (tc:applied-type-expr->operand-types x)
			 (tc:applied-type-expr->operand-types y)
			 bindings
			 var-list)))
     ((tc:and-type-expr? x)
      (tc:find-and-solution x y bindings var-list))
     ((tc:and-type-expr? y)
      (tc:find-and-solution y x bindings var-list))
     ((and (tc:varargs-function-type-expr? x)
	   (tc:function-type-expr? y))
      (tc:unify (tc:fix-length x y) y bindings var-list))
     ((and (tc:varargs-function-type-expr? y)
	   (tc:function-type-expr? x))
      (tc:unify x (tc:fix-length y x) bindings var-list))
     ((and (tc:function-type-expr? x) (tc:function-type-expr? y))
      (let ((args-res (tc:list-unify
		       (tc:function-type-expr->arg-types x)
		       (tc:function-type-expr->arg-types y)
		       bindings
		       var-list)))
	(if (tc:unification-error? args-res)
	    args-res
	    (tc:unify (tc:function-type-expr->result-type x)
		      (tc:function-type-expr->result-type y)
		      (something->value args-res)
		      var-list))))
     (else *tc:unification-failure*))))

(define tc:list-unify
  ;; TYPE: (-> ((list type-expr)
  ;;            (list type-expr)
  ;;            (env uniq-sym type-expr)
  ;;            (list variable-type-expr))
  ;;           (maybe (env uniq-sym type-expr)))
  (lambda (ls1 ls2 bindings var-list)
    ;; (displayln "tc:list-unify " ls1 " and " ls2)
    (cond
     ((null? ls1)
      (if (null? ls2)
	  (make-something bindings)
	  *tc:unification-failure*))
     ((null? ls2) *tc:unification-failure*)
     (else
      (maybe-test (tc:unify (car ls1) (car ls2) bindings var-list)
		  (lambda (new-bindings)
		    (tc:list-unify (cdr ls1) (cdr ls2) new-bindings var-list))
		  (lambda () *tc:unification-failure*))))))
	
(define tc:unify-variable 
  ;; TYPE: (-> (uniq-sym
  ;;            type-expr
  ;;            (env uniq-sym type-expr)
  ;;            (list variable-type-expr))
  ;;           (maybe (env uniq-sym type-expr)))
  (lambda (var x bindings var-list)
    ;; ENSURES: result is either *tc:unification-failure*
    ;;          or the unifing env for var with x, which may be an extension
    ;;          of bindings.
    ;; Note especially the handling of poof and datum.
    (if (>= *tc:tracing-level* 2)
	(begin
	  (displayln "tc:unify-variable " var " and " x)
	  (displayln "tc:unify-variable bindings = " bindings)))
    (cond
     ((tc:env-bound? var bindings)
      (tc:unify (tc:env-value var bindings) x bindings var-list))
     ((and (tc:variable-type-expr? x)
	   (tc:env-bound? (tc:variable-type-expr->uniq-sym x) bindings))
      (tc:unify (tc:uniq-sym->variable-type-expr var)
		(tc:env-value (tc:variable-type-expr->uniq-sym x) bindings)
		bindings
		var-list))
     ((tc:occurs-check? var x bindings) *tc:unification-failure*)
     ((tc:poof-type-expr? x)
      (make-something bindings))
     ((tc:datum-type-expr? x)
      (make-something (tc:env-update var x bindings)))
     ((tc:variable-type-expr? x)
      (if (tc:member? x var-list)
	  (make-something (tc:env-update var x bindings))
	  (make-something (tc:env-update
			   (tc:variable-type-expr->uniq-sym x)
			   (tc:uniq-sym->variable-type-expr var)
			   bindings))))
     (else (make-something (tc:env-update var x bindings))))))

(define tc:occurs-check?
  ;; TYPE: (-> (uniq-sym type-expr (env uniq-sym type-expr)) boolean)
  (lambda (var x bindings)
    ;; ENSURES: result is #t iff var occurs anywhere inside x
    (or (and (tc:variable-type-expr? x)
	     (let ((x-as-us (tc:variable-type-expr->uniq-sym x)))
	       (or (eq? var x-as-us)
		   (and (tc:env-bound? x-as-us bindings)
			(tc:occurs-check? var
					  (tc:env-value x-as-us bindings)
					  bindings)))))
	(and (tc:function-type-expr? x)
	     (or (tc:occurs-check? var
				   (tc:function-type-expr->result-type x)
				   bindings)
		 (tc:some (lambda (t)
			    (tc:occurs-check? var t bindings))
			  (tc:function-type-expr->arg-types x))))
	(and (tc:applied-type-expr? x)
	     (tc:some (lambda (t)
			(tc:occurs-check? var t bindings))
		      (tc:applied-type-expr->operand-types x)))
	(and (tc:and-type-expr? x)
	     (tc:some (lambda (t)
			(tc:occurs-check? var t bindings))
		      (tc:and-type-expr->conjoined-types x))))))

(define tc:find-and-solution
  ;; TYPE: (-> (type-expr
  ;;            type-expr
  ;;            (env uniq-sym type-expr)
  ;;            (list variable-type-expr))
  ;;           (maybe (env uniq-sym type-expr)))
  (lambda (conj-type sought e var-list)
    ;; EFFECT: attempts to unify "sought" with each conjoined type in conj-type
    ;;         in sequential order returning the first unifying environment
    ;;         or *tc:unification-failure* if no such environment exists
    (letrec
	((find-solution
	  ;; TYPE: (-> ((list type-expr)) (maybe (env uniq-sym type-expr)))
	  (lambda (type-list)
	    (if (null? type-list)
		*tc:unification-failure*
		(let ((unification-result
		       (tc:unify (car type-list) sought e var-list)))
		  (if (not (tc:unification-error? unification-result))
		      unification-result
		      (find-solution (cdr type-list))))))))
      (find-solution (tc:and-type-expr->conjoined-types conj-type)))))


(define tc:fix-length
  ;; TYPE: (-> (type-expr type-expr) type-expr)
  (letrec ((get-dot-type
	    ;; TYPE: (-> ((list type-expr)) type-expr)
	    (lambda (ls)
	      ;; REQUIRES: ls at least one type followed by '...' in it
	      ;; EFFECT: returns the element of the ls before the '...'
	      (if (tc:dots-type-expr? (cadr ls))
		  (car ls)
		  (get-dot-type (cdr ls)))))
	   (preceed-dot-type;; TYPE: (-> ((list type-expr)) (list type-expr))
	    (lambda (ls)
	      ;; REQUIRES: ls at least one type followed by '...' in it
	      ;; EFFECT: returns the part of the list before the '...'
	      ;;         not including the type just before the '...'
	      (cond
	       ((null? (cddr ls)) '())
	       ((tc:dots-type-expr? (caddr ls)) (list (car ls)))
	       (else (cons (car ls) (preceed-dot-type (cdr ls)))))))
	   (enough-copies
	    ;; TYPE: (-> (number type-expr) (list type-expr))
	    (lambda (n dotted)
	      (if (<= n 0)
		  '()
		  (cons dotted (enough-copies (- n 1) dotted))))))
    (lambda (dot-type arrow-type)
      ;; REQUIRES: dot-type is an arrow type with '...' in it
      ;; EFFECT: result is a type like dot-type but with an appropriate
      ;;         number of copies of types to fill out the ... in hopes of
      ;;         matching arrow-type
      (let ((type-list (tc:function-type-expr->arg-types dot-type))
	    (arg-list (tc:function-type-expr->arg-types arrow-type)))
	(tc:make-function-type-expr
	 (let* ((dot-type (get-dot-type type-list))
		(before (preceed-dot-type type-list)))
	   (append
	    before
	    (enough-copies (- (length arg-list) (length before))
			   dot-type)))
	 (tc:function-type-expr->result-type dot-type))))))

(define tc:subst-bindings-c
  ;; TYPE: (-> ((env uniq-sym type-expr)) (-> (type-expr) type-expr)
  (lambda (env)
    (letrec ((resproc 
	      (lambda (x)
		;; TYPE: (-> (type-expr) type-expr)
		;; EFFECT: returns a type-expr like x but with each variable 
		;;         bound in env replaced by its binding and 
		;;         recursively bound variables accounted for
		(cond
		 ((and (tc:variable-type-expr? x)
		       (tc:env-bound? (tc:variable-type-expr->uniq-sym x)
				      env))
		  (resproc (tc:env-value (tc:variable-type-expr->uniq-sym x)
					  env)))
		 ((or (tc:variable-type-expr? x) (tc:basic-type-expr? x)) x)
		 ((tc:function-type-expr? x)
		  (tc:make-function-type-expr
		   (map (lambda (t) (resproc t))
			(tc:function-type-expr->arg-types x))
		   (resproc (tc:function-type-expr->result-type x))))
		 ((tc:and-type-expr? x)
		  (tc:make-and-type-expr
		   (map (lambda (t) (resproc t))
			(tc:and-type-expr->conjoined-types x))))
		 ((tc:applied-type-expr? x)
		  (tc:make-applied-type-expr
		   (tc:applied-type-expr->operator-type x)
		   (map (lambda (t) (resproc t))
			(tc:applied-type-expr->operand-types x))))
		 (else
		  (error "tc:subst-bindings-c, illegal type syntax: " x))))))
      resproc)))
      
(define tc:subst-bindings
  ;; TYPE: (-> ((env uniq-sym type-expr) type-expr) type-expr)
  (lambda (bindings x)
    ;; EFFECT: returns a type-expr like x but with each variable 
    ;;         bound in env replaced by its binding and 
    ;;         recursively bound variables accounted for
    ;; (displayln "tc:subst-bindings with bindings " bindings " and x = " x)
    ((tc:subst-bindings-c bindings) x)))

(define tc:unifier 
  ;; TYPE: (-> (type-expr type-expr) (maybe type-expr))
  (lambda (x y)
    ;; ENSURES: result is either (make-nothing) or
    ;;          a type-expression that is the application of
    ;;          the unifying substitution to the types
    (let ((unif-res (tc:unify x y (tc:env-empty) '())))
      (if (tc:unification-error? unif-res)
	  (make-nothing)
	  (make-something
	   (tc:subst-bindings (something->value unif-res) x))))))

